@@ -92,11 +92,9 @@ struct _GabbleMediaSessionPrivate
   gchar *object_path;
 
   GPtrArray *streams;
-  GHashTable *streams_by_name;
   GPtrArray *remove_requests;
 
   gchar *id;
-  JingleInitiator initiator;
   GabbleHandle peer;
   gchar *peer_resource;
 
@@ -139,8 +137,6 @@ gabble_media_session_init (GabbleMediaSession *self)
   priv->mode = MODE_JINGLE;
   priv->state = JS_STATE_PENDING_CREATED;
   priv->streams = g_ptr_array_new ();
-  priv->streams_by_name = g_hash_table_new_full (g_str_hash, g_str_equal,
-      g_free, g_object_unref);
   priv->remove_requests = g_ptr_array_new ();
 }
 
@@ -172,6 +168,33 @@ _emit_new_stream (GabbleMediaSession *session,
   g_free (object_path);
 }
 
+
+static GabbleMediaStream *
+_lookup_stream_by_name_and_initiator (GabbleMediaSession *session,
+                                      const gchar *stream_name,
+                                      JingleInitiator stream_initiator)
+{
+  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
+  guint i;
+
+  for (i = 0; i < priv->streams->len; i++)
+    {
+      GabbleMediaStream *stream = g_ptr_array_index (priv->streams, i);
+
+      if (g_strdiff (stream->name, stream_name))
+        continue;
+
+      if (stream_initiator != INITIATOR_INVALID &&
+          stream_initiator != stream->initiator)
+        continue;
+
+      return stream;
+    }
+
+  return NULL;
+}
+
+
 static GabbleMediaStream *
 create_media_stream (GabbleMediaSession *session,
                      const gchar *name,
@@ -200,7 +223,8 @@ create_media_stream (GabbleMediaSession *session,
     }
 
   g_assert (priv->streams->len < MAX_STREAMS);
-  g_assert (g_hash_table_lookup (priv->streams_by_name, name) == NULL);
+  g_assert (_lookup_stream_by_name_and_initiator (session, name, initiator) ==
+      NULL);
 
   id = _gabble_media_channel_get_stream_id (priv->channel);
 
@@ -231,7 +255,6 @@ create_media_stream (GabbleMediaSession *session,
                     session);
 
   g_ptr_array_add (priv->streams, stream);
-  g_hash_table_insert (priv->streams_by_name, g_strdup (name), stream);
 
   g_free (object_path);
 
@@ -245,13 +268,13 @@ create_media_stream (GabbleMediaSession *session,
 
 static void
 destroy_media_stream (GabbleMediaSession *session,
-                     GabbleMediaStream *stream)
+                      GabbleMediaStream *stream)
 {
   GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
 
   _gabble_media_stream_close (stream);
   g_ptr_array_remove_fast (priv->streams, stream);
-  g_hash_table_remove (priv->streams_by_name, stream->name);
+  g_object_unref (stream);
 }
 
 static GObject *
@@ -295,7 +318,7 @@ gabble_media_session_get_property (GObject    *object,
       g_value_set_string (value, priv->id);
       break;
     case PROP_INITIATOR:
-      g_value_set_uint (value, priv->initiator);
+      g_value_set_uint (value, session->initiator);
       break;
     case PROP_PEER:
       g_value_set_uint (value, priv->peer);
@@ -342,7 +365,7 @@ gabble_media_session_set_property (GObject      *object,
       priv->id = g_value_dup_string (value);
       break;
     case PROP_INITIATOR:
-      priv->initiator = g_value_get_uint (value);
+      session->initiator = g_value_get_uint (value);
       break;
     case PROP_PEER:
       priv->peer = g_value_get_uint (value);
@@ -522,13 +545,12 @@ gabble_media_session_dispose (GObject *object)
 
   if (priv->streams != NULL)
     {
+      for (i = 0; i < priv->streams->len; i++)
+        g_object_unref (g_ptr_array_index (priv->streams, i));
       g_ptr_array_free (priv->streams, TRUE);
       priv->streams = NULL;
     }
 
-  g_hash_table_destroy (priv->streams_by_name);
-  priv->streams_by_name = NULL;
-
   for (i = 0; i < priv->remove_requests->len; i++)
     g_ptr_array_free (g_ptr_array_index (priv->remove_requests, i), TRUE);
   g_ptr_array_free (priv->remove_requests, TRUE);
@@ -641,22 +663,6 @@ gabble_media_session_ready (GabbleMediaSession *self,
   return TRUE;
 }
 
-static GabbleMediaStream *
-_lookup_stream_by_name (GabbleMediaSession *session,
-                        const gchar *stream_name)
-{
-  GabbleMediaSessionPrivate *priv;
-  GabbleMediaStream *stream;
-
-  g_assert (GABBLE_IS_MEDIA_SESSION (session));
-
-  priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
-
-  stream = g_hash_table_lookup (priv->streams_by_name, stream_name);
-
-  return stream;
-}
-
 
 static gboolean
 _handle_create (GabbleMediaSession *session,
@@ -674,7 +680,7 @@ _handle_create (GabbleMediaSession *session,
   gboolean override_existing = FALSE;
 
   if ((priv->state == JS_STATE_PENDING_CREATED) &&
-      (priv->initiator == INITIATOR_LOCAL))
+      (session->initiator == INITIATOR_LOCAL))
     {
       DEBUG ("we're trying to call ourselves, rejecting with busy");
       _gabble_media_session_terminate (session, INITIATOR_REMOTE,
@@ -689,7 +695,7 @@ _handle_create (GabbleMediaSession *session,
        * streams which we are trying to add (but havn't had acknowledged) */
       if (stream->signalling_state < STREAM_SIG_STATE_ACKNOWLEDGED)
         {
-          if (priv->initiator == INITIATOR_REMOTE)
+          if (session->initiator == INITIATOR_REMOTE)
             {
               override_existing = TRUE;
             }
@@ -810,19 +816,18 @@ static TpMediaStreamDirection
 _senders_to_direction (GabbleMediaSession *session,
                        const gchar *senders)
 {
-  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   TpMediaStreamDirection ret = TP_MEDIA_STREAM_DIRECTION_NONE;
 
   if (!g_strdiff (senders, "initiator"))
     {
-      if (priv->initiator == INITIATOR_LOCAL)
+      if (session->initiator == INITIATOR_LOCAL)
         ret = TP_MEDIA_STREAM_DIRECTION_SEND;
       else
         ret = TP_MEDIA_STREAM_DIRECTION_RECEIVE;
     }
   else if (!g_strdiff (senders, "responder"))
     {
-      if (priv->initiator == INITIATOR_REMOTE)
+      if (session->initiator == INITIATOR_REMOTE)
         ret = TP_MEDIA_STREAM_DIRECTION_SEND;
       else
         ret = TP_MEDIA_STREAM_DIRECTION_RECEIVE;
@@ -897,7 +902,10 @@ _handle_direction (GabbleMediaSession *session,
   /* make any necessary changes */
   new_combined_dir = MAKE_COMBINED_DIRECTION (requested_dir, pending_send);
   if (new_combined_dir != stream->combined_direction)
-    g_object_set (stream, "combined-direction", new_combined_dir, NULL);
+    {
+      g_object_set (stream, "combined-direction", new_combined_dir, NULL);
+      _gabble_media_stream_update_sending (stream, FALSE);
+    }
 
   return TRUE;
 }
@@ -915,6 +923,8 @@ _handle_accept (GabbleMediaSession *session,
 {
   g_object_set (stream, "playing", TRUE, NULL);
 
+  _gabble_media_stream_update_sending (stream, TRUE);
+
   return TRUE;
 }
 
@@ -1137,6 +1147,7 @@ _call_handlers_on_stream (GabbleMediaSession *session,
                           LmMessage *message,
                           LmMessageNode *content_node,
                           const gchar *stream_name,
+                          JingleInitiator stream_creator,
                           StreamHandlerFunc *func,
                           GError **error)
 {
@@ -1157,7 +1168,8 @@ _call_handlers_on_stream (GabbleMediaSession *session,
     {
       /* handlers may create the stream */
       if (stream == NULL && stream_name != NULL)
-        stream = _lookup_stream_by_name (session, stream_name);
+        stream = _lookup_stream_by_name_and_initiator (session, stream_name,
+            stream_creator);
 
       /* the create handler is able to check whether or not the stream
        * exists, and act accordingly (sometimes it will replace an existing
@@ -1168,9 +1180,17 @@ _call_handlers_on_stream (GabbleMediaSession *session,
           /* all other handlers require the stream to exist */
           if (stream == NULL)
             {
+              const gchar *created = "";
+
+              if (stream_creator == INITIATOR_LOCAL)
+                created = "locally-created ";
+              else if (stream_creator == INITIATOR_REMOTE)
+                created = "remotely-created ";
+
               g_set_error (error, GABBLE_XMPP_ERROR, XMPP_ERROR_ITEM_NOT_FOUND,
-                  "unable to handle action for unknown stream \"%s\"",
-                  stream_name);
+                  "unable to handle action for unknown %sstream \"%s\" ",
+                  created, stream_name);
+
               return FALSE;
             }
           else
@@ -1214,6 +1234,28 @@ _call_handlers_on_stream (GabbleMediaSession *session,
 }
 
 
+static JingleInitiator
+_creator_to_initiator (GabbleMediaSession *session, const gchar *creator)
+{
+  if (!g_strdiff (creator, "initiator"))
+    {
+      if (session->initiator == INITIATOR_LOCAL)
+        return INITIATOR_LOCAL;
+      else
+        return INITIATOR_REMOTE;
+    }
+  else if (!g_strdiff (creator, "responder"))
+    {
+      if (session->initiator == INITIATOR_LOCAL)
+        return INITIATOR_REMOTE;
+      else
+        return INITIATOR_LOCAL;
+    }
+  else
+    return INITIATOR_INVALID;
+}
+
+
 static gboolean
 _call_handlers_on_streams (GabbleMediaSession *session,
                            LmMessage *message,
@@ -1221,42 +1263,50 @@ _call_handlers_on_streams (GabbleMediaSession *session,
                            StreamHandlerFunc *func,
                            GError **error)
 {
+  LmMessageNode *content_node;
+
   if (lm_message_node_has_namespace (session_node, NS_GOOGLE_SESSION, NULL))
-    {
-      if (!_call_handlers_on_stream (session, message, session_node,
-            GTALK_STREAM_NAME, func, error))
-        return FALSE;
-    }
-  else
-    {
-      LmMessageNode *content_node;
+    return _call_handlers_on_stream (session, message, session_node,
+        GTALK_STREAM_NAME, INITIATOR_INVALID, func, error);
 
-      if (session_node->children == NULL)
-        return _call_handlers_on_stream (session, message, NULL, NULL, func,
-            error);
+  if (session_node->children == NULL)
+    return _call_handlers_on_stream (session, message, NULL, NULL,
+        INITIATOR_INVALID, func, error);
 
-      for (content_node = session_node->children;
-           NULL != content_node;
-           content_node = content_node->next)
+  for (content_node = session_node->children;
+       NULL != content_node;
+       content_node = content_node->next)
+    {
+      const gchar *stream_name, *stream_creator;
+      JingleInitiator stream_initiator;
+
+      if (g_strdiff (content_node->name, "content"))
+        continue;
+
+      stream_name = lm_message_node_get_attribute (content_node, "name");
+
+      if (stream_name == NULL)
         {
-          const gchar *stream_name;
-
-          if (g_strdiff (content_node->name, "content"))
-            continue;
-
-          stream_name = lm_message_node_get_attribute (content_node, "name");
-
-          if (stream_name == NULL)
-            {
-              g_set_error (error, GABBLE_XMPP_ERROR, XMPP_ERROR_BAD_REQUEST,
-                  "rejecting content node with no name");
-              return FALSE;
-            }
-
-          if (!_call_handlers_on_stream (session, message, content_node,
-                stream_name, func, error))
-            return FALSE;
+          g_set_error (error, GABBLE_XMPP_ERROR, XMPP_ERROR_BAD_REQUEST,
+              "rejecting content node with no name");
+          return FALSE;
         }
+
+      stream_creator = lm_message_node_get_attribute (content_node, "creator");
+      stream_initiator = _creator_to_initiator (session, stream_creator);
+
+      /* we allow NULL creator to mean INITIATOR_INVALID for backwards
+       * compatibility with clients that don't put a creator attribute in */
+      if (stream_creator != NULL && stream_initiator == INITIATOR_INVALID)
+        {
+          g_set_error (error, GABBLE_XMPP_ERROR, XMPP_ERROR_BAD_REQUEST,
+              "rejecting content node with invalid creators value");
+          return FALSE;
+        }
+
+      if (!_call_handlers_on_stream (session, message, content_node,
+            stream_name, stream_initiator, func, error))
+        return FALSE;
     }
 
   return TRUE;
@@ -1493,46 +1543,28 @@ _set_remote_streams_playing (GabbleMediaSession *session)
     }
 }
 
-typedef struct _AddDescriptionsData AddDescriptionsData;
-
-struct _AddDescriptionsData {
-  GabbleMediaSession *session;
-  LmMessageNode *session_node;
-  JingleInitiator initiator;
-};
-
 static const gchar *_direction_to_senders (GabbleMediaSession *,
     TpMediaStreamDirection);
 
 static void
-_add_content_descriptions_one (const gchar *name,
+_add_content_descriptions_one (GabbleMediaSession *session,
                                GabbleMediaStream *stream,
-                               AddDescriptionsData *data)
+                               LmMessageNode *session_node)
 {
-  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (data->session);
+  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   LmMessageNode *content_node;
 
-  if (stream->initiator != data->initiator)
-    {
-      GMS_DEBUG_INFO (data->session,
-          "not adding content description for %s stream %s",
-          stream->initiator == INITIATOR_LOCAL ? "local" : "remote",
-          name);
-      return;
-    }
-
   if (priv->mode == MODE_GOOGLE)
     {
-      content_node = data->session_node;
+      content_node = session_node;
     }
   else
     {
       TpMediaStreamDirection direction;
       TpMediaStreamPendingSend pending_send;
 
-      content_node = lm_message_node_add_child (data->session_node, "content",
-          NULL);
-      lm_message_node_set_attribute (content_node, "name", name);
+      content_node = _gabble_media_stream_add_content_node (stream,
+          session_node);
 
       direction = COMBINED_DIRECTION_GET_DIRECTION (stream->combined_direction);
       pending_send = COMBINED_DIRECTION_GET_PENDING_SEND
@@ -1550,7 +1582,7 @@ _add_content_descriptions_one (const gchar *name,
       if (direction != TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL)
         {
           const gchar *senders;
-          senders = _direction_to_senders (data->session, direction);
+          senders = _direction_to_senders (session, direction);
           lm_message_node_set_attribute (content_node, "senders", senders);
         }
     }
@@ -1566,14 +1598,23 @@ _add_content_descriptions (GabbleMediaSession *session,
                            JingleInitiator stream_initiator)
 {
   GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
-  AddDescriptionsData data;
+  guint i;
 
-  data.session = session;
-  data.session_node = session_node;
-  data.initiator = stream_initiator;
+  for (i = 0; i < priv->streams->len; i++)
+    {
+      GabbleMediaStream *stream = g_ptr_array_index (priv->streams, i);
 
-  g_hash_table_foreach (priv->streams_by_name,
-      (GHFunc) _add_content_descriptions_one, &data);
+      if (stream->initiator != stream_initiator)
+        {
+          GMS_DEBUG_INFO (session,
+              "not adding content description for %s stream %s",
+              stream->initiator == INITIATOR_LOCAL ? "local" : "remote",
+              stream->name);
+          continue;
+        }
+
+      _add_content_descriptions_one (session, stream, session_node);
+    }
 }
 
 static LmHandlerResult
@@ -1584,18 +1625,29 @@ accept_msg_reply_cb (GabbleConnection *conn,
                      gpointer user_data)
 {
   GabbleMediaSession *session = GABBLE_MEDIA_SESSION (object);
+  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
+  guint i;
 
   MSG_REPLY_CB_END_SESSION_IF_NOT_SUCCESSFUL (session, "accept failed");
 
+  for (i = 0; i < priv->streams->len; i++)
+    {
+      GabbleMediaStream *stream = g_ptr_array_index (priv->streams, i);
+
+      if (stream->initiator == INITIATOR_LOCAL)
+        continue;
+
+      _gabble_media_stream_update_sending (stream, TRUE);
+    }
+
   g_object_set (session, "state", JS_STATE_ACTIVE, NULL);
 
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
 static gboolean
-_stream_not_ready_for_accept (const gchar *name,
-                              GabbleMediaStream *stream,
-                              GabbleMediaSession *session)
+_stream_not_ready_for_accept (GabbleMediaSession *session,
+                              GabbleMediaStream *stream)
 {
   /* locally initiated streams shouldn't delay acceptance */
   if (stream->initiator == INITIATOR_LOCAL)
@@ -1604,14 +1656,14 @@ _stream_not_ready_for_accept (const gchar *name,
   if (!stream->got_local_codecs)
     {
       GMS_DEBUG_INFO (session, "stream %s does not yet have local codecs",
-          name);
+          stream->name);
 
       return TRUE;
     }
 
   if (stream->connection_state != TP_MEDIA_STREAM_STATE_CONNECTED)
     {
-      GMS_DEBUG_INFO (session, "stream %s is not yet connected", name);
+      GMS_DEBUG_INFO (session, "stream %s is not yet connected", stream->name);
 
       return TRUE;
     }
@@ -1626,6 +1678,7 @@ try_session_accept (GabbleMediaSession *session)
   LmMessage *msg;
   LmMessageNode *session_node;
   const gchar *action;
+  guint i;
 
   if (priv->state < JS_STATE_ACTIVE && !priv->locally_accepted)
     {
@@ -1634,12 +1687,16 @@ try_session_accept (GabbleMediaSession *session)
       return;
     }
 
-  if (g_hash_table_find (priv->streams_by_name,
-        (GHRFunc) _stream_not_ready_for_accept, session) != NULL)
+  for (i = 0; i < priv->streams->len; i++)
     {
-      GMS_DEBUG_INFO (session, "not sending accept yet, found a stream which "
-          "was not yet connected or was missing local codecs");
-      return;
+      GabbleMediaStream *stream = g_ptr_array_index (priv->streams, i);
+
+      if (_stream_not_ready_for_accept (session, stream))
+        {
+          GMS_DEBUG_INFO (session, "not sending accept yet, found a stream "
+              "which was not yet connected or was missing local codecs");
+          return;
+        }
     }
 
   if (priv->mode == MODE_GOOGLE)
@@ -1690,6 +1747,8 @@ content_accept_msg_reply_cb (GabbleConnection *conn,
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
     }
 
+  _gabble_media_stream_update_sending (stream, TRUE);
+
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
@@ -1700,12 +1759,11 @@ try_content_accept (GabbleMediaSession *session,
   GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   LmMessage *msg;
   LmMessageNode *session_node;
-  AddDescriptionsData data;
 
   g_assert (priv->state == JS_STATE_ACTIVE);
   g_assert (priv->mode == MODE_JINGLE);
 
-  if (_stream_not_ready_for_accept (stream->name, stream, session))
+  if (_stream_not_ready_for_accept (session, stream))
     {
       GMS_DEBUG_INFO (session, "not sending content-accept yet, stream %s "
           "is disconnected or missing local codecs", stream->name);
@@ -1716,11 +1774,7 @@ try_content_accept (GabbleMediaSession *session,
   msg = _gabble_media_session_message_new (session, "content-accept",
       &session_node);
 
-  data.session = session;
-  data.session_node = session_node;
-  data.initiator = INITIATOR_REMOTE;
-
-  _add_content_descriptions_one (stream->name, stream, &data);
+  _add_content_descriptions_one (session, stream, session_node);
 
   GMS_DEBUG_INFO (session, "sending jingle session action \"content-accept\" "
       "to peer for stream %s", stream->name);
@@ -1754,14 +1808,13 @@ initiate_msg_reply_cb (GabbleConnection *conn,
 }
 
 static gboolean
-_stream_not_ready_for_initiate (const gchar *name,
-                                GabbleMediaStream *stream,
-                                GabbleMediaSession *session)
+_stream_not_ready_for_initiate (GabbleMediaSession *session,
+                                GabbleMediaStream *stream)
 {
   if (!stream->got_local_codecs)
     {
       GMS_DEBUG_INFO (session, "stream %s does not yet have local codecs",
-          name);
+          stream->name);
 
       return TRUE;
     }
@@ -1776,13 +1829,18 @@ try_session_initiate (GabbleMediaSession *session)
   LmMessage *msg;
   LmMessageNode *session_node;
   const gchar *action;
+  guint i;
 
-  if (g_hash_table_find (priv->streams_by_name,
-        (GHRFunc) _stream_not_ready_for_initiate, session) != NULL)
+  for (i = 0; i < priv->streams->len; i++)
     {
-      GMS_DEBUG_INFO (session, "not sending initiate yet, found a stream "
-          "which was missing local codecs");
-      return;
+      GabbleMediaStream *stream = g_ptr_array_index (priv->streams, i);
+
+      if (_stream_not_ready_for_initiate (session, stream))
+        {
+          GMS_DEBUG_INFO (session, "not sending initiate yet, found a stream "
+            "which was missing local codecs");
+          return;
+        }
     }
 
   if (priv->mode == MODE_GOOGLE)
@@ -1815,12 +1873,11 @@ content_add_msg_reply_cb (GabbleConnection *conn,
                           gpointer user_data)
 {
   GabbleMediaSession *session = GABBLE_MEDIA_SESSION (user_data);
-  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   GabbleMediaStream *stream = GABBLE_MEDIA_STREAM (object);
 
   if (lm_message_get_sub_type (reply_msg) != LM_MESSAGE_SUB_TYPE_RESULT)
     {
-      if (priv->initiator == INITIATOR_REMOTE &&
+      if (session->initiator == INITIATOR_REMOTE &&
           stream->signalling_state == STREAM_SIG_STATE_ACKNOWLEDGED)
         {
           GMS_DEBUG_INFO (session, "ignoring content-add failure, stream has "
@@ -1864,12 +1921,11 @@ do_content_add (GabbleMediaSession *session,
   GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   LmMessage *msg;
   LmMessageNode *session_node;
-  AddDescriptionsData data;
 
   g_assert (priv->state == JS_STATE_ACTIVE);
   g_assert (priv->mode == MODE_JINGLE);
 
-  if (_stream_not_ready_for_initiate (stream->name, stream, session))
+  if (_stream_not_ready_for_initiate (session, stream))
     {
       GMS_DEBUG_ERROR (session, "trying to send content-add for stream %s "
           "but we have no local codecs. what?!", stream->name);
@@ -1880,11 +1936,7 @@ do_content_add (GabbleMediaSession *session,
   msg = _gabble_media_session_message_new (session, "content-add",
       &session_node);
 
-  data.session = session;
-  data.session_node = session_node;
-  data.initiator = INITIATOR_LOCAL;
-
-  _add_content_descriptions_one (stream->name, stream, &data);
+  _add_content_descriptions_one (session, stream, session_node);
 
   GMS_DEBUG_INFO (session, "sending jingle action \"content-add\" to peer for "
       "stream %s", stream->name);
@@ -1924,7 +1976,7 @@ stream_connection_state_changed_cb (GabbleMediaStream *stream,
   if (priv->state < JS_STATE_ACTIVE)
     {
       /* send a session accept if the session was initiated by the peer */
-      if (priv->initiator == INITIATOR_REMOTE)
+      if (session->initiator == INITIATOR_REMOTE)
         {
           try_session_accept (session);
         }
@@ -1976,7 +2028,7 @@ stream_got_local_codecs_changed_cb (GabbleMediaStream *stream,
   /* after session is active, we do things per-stream with content-* actions */
   if (priv->state < JS_STATE_ACTIVE)
     {
-      if (priv->initiator == INITIATOR_REMOTE)
+      if (session->initiator == INITIATOR_REMOTE)
         {
           if (priv->state < JS_STATE_PENDING_ACCEPT_SENT)
             {
@@ -2079,7 +2131,7 @@ _gabble_media_session_message_new (GabbleMediaSession *session,
   else
     element = "jingle";
 
-  if (priv->initiator == INITIATOR_LOCAL)
+  if (session->initiator == INITIATOR_LOCAL)
     initiator_handle = priv->conn->self_handle;
   else
     initiator_handle = priv->peer;
@@ -2135,6 +2187,7 @@ _gabble_media_session_accept (GabbleMediaSession *session)
           pending_send &= ~TP_MEDIA_STREAM_PENDING_LOCAL_SEND;
           combined_dir = MAKE_COMBINED_DIRECTION (current_dir, pending_send);
           g_object_set (stream, "combined-direction", combined_dir, NULL);
+          _gabble_media_stream_update_sending (stream, FALSE);
         }
     }
 
@@ -2215,9 +2268,8 @@ _gabble_media_session_remove_streams (GabbleMediaSession *session,
             g_assert (msg != NULL);
             g_assert (removing != NULL);
 
-            content_node = lm_message_node_add_child (session_node, "content",
-              NULL);
-            lm_message_node_set_attribute (content_node, "name", stream->name);
+            content_node = _gabble_media_stream_add_content_node (stream,
+                session_node);
 
             g_object_set (stream,
                 "playing", FALSE,
@@ -2351,7 +2403,7 @@ _gabble_media_session_terminate (GabbleMediaSession *session,
       /* Jingle doesn't have a "reject" action; a termination before an
        * acceptance indicates that the call has been declined */
 
-      if (priv->initiator == INITIATOR_REMOTE &&
+      if (session->initiator == INITIATOR_REMOTE &&
           priv->state == JS_STATE_PENDING_INITIATED &&
           priv->mode == MODE_GOOGLE)
         {
@@ -2459,7 +2511,11 @@ _name_stream (GabbleMediaSession *session,
               media_type == TP_MEDIA_STREAM_TYPE_AUDIO ? "audio" : "video",
               i++);
 
-          if (g_hash_table_lookup (priv->streams_by_name, ret) != NULL)
+          /* even though we now have seperate namespaces for local and remote,
+           * actually check in both so that we can still support clients which
+           * have 1 namespace (such as our older selves :D) */
+          if (_lookup_stream_by_name_and_initiator (session, ret,
+                INITIATOR_INVALID) != NULL)
             {
               ret[0] = '\0';
             }
@@ -2685,7 +2741,6 @@ static const gchar *
 _direction_to_senders (GabbleMediaSession *session,
                        TpMediaStreamDirection dir)
 {
-  GabbleMediaSessionPrivate *priv = GABBLE_MEDIA_SESSION_GET_PRIVATE (session);
   const gchar *ret = NULL;
 
   switch (dir)
@@ -2694,13 +2749,13 @@ _direction_to_senders (GabbleMediaSession *session,
         g_assert_not_reached ();
         break;
       case TP_MEDIA_STREAM_DIRECTION_SEND:
-        if (priv->initiator == INITIATOR_LOCAL)
+        if (session->initiator == INITIATOR_LOCAL)
           ret = "initiator";
         else
           ret = "responder";
         break;
       case TP_MEDIA_STREAM_DIRECTION_RECEIVE:
-        if (priv->initiator == INITIATOR_REMOTE)
+        if (session->initiator == INITIATOR_REMOTE)
           ret = "initiator";
         else
           ret = "responder";
@@ -2722,10 +2777,16 @@ direction_msg_reply_cb (GabbleConnection *conn,
                         GObject *object,
                         gpointer user_data)
 {
-  GabbleMediaSession *session = GABBLE_MEDIA_SESSION (object);
+  GabbleMediaSession *session = GABBLE_MEDIA_SESSION (user_data);
+  GabbleMediaStream *stream = GABBLE_MEDIA_STREAM (object);
 
   MSG_REPLY_CB_END_SESSION_IF_NOT_SUCCESSFUL (session, "direction change failed");
 
+  if (stream->playing)
+    {
+      _gabble_media_stream_update_sending (stream, TRUE);
+    }
+
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
@@ -2758,15 +2819,12 @@ send_direction_change (GabbleMediaSession *session,
 
   msg = _gabble_media_session_message_new (session, "content-modify",
       &session_node);
-  content_node = lm_message_node_add_child (session_node, "content", NULL);
+  content_node = _gabble_media_stream_add_content_node (stream, session_node);
 
-  lm_message_node_set_attributes (content_node,
-      "name", stream->name,
-      "senders", senders,
-      NULL);
+  lm_message_node_set_attribute (content_node, "senders", senders);
 
   ret = _gabble_connection_send_with_reply (priv->conn, msg,
-      direction_msg_reply_cb, G_OBJECT (session), NULL, error);
+      direction_msg_reply_cb, G_OBJECT (stream), session, error);
 
   lm_message_unref (msg);
 
@@ -2837,7 +2895,10 @@ _gabble_media_session_request_stream_direction (GabbleMediaSession *session,
   /* make any necessary changes */
   new_combined_dir = MAKE_COMBINED_DIRECTION (requested_dir, pending_send);
   if (new_combined_dir != stream->combined_direction)
-    g_object_set (stream, "combined-direction", new_combined_dir, NULL);
+    {
+      g_object_set (stream, "combined-direction", new_combined_dir, NULL);
+      _gabble_media_stream_update_sending (stream, FALSE);
+    }
 
   /* short-circuit sending a request if we're not asking for anything new */
   if (current_dir == requested_dir)
