@@ -38,6 +38,7 @@
 
 #include "telepathy-helpers.h"
 #include "telepathy-constants.h"
+#include "telepathy-errors.h"
 
 #include "gabble-media-stream.h"
 #include "gabble-media-stream-signals-marshal.h"
@@ -309,17 +310,7 @@ gabble_media_stream_set_property (GObject      *object,
         }
       break;
     case PROP_COMBINED_DIRECTION:
-        {
-          gboolean new_sending;
-          stream->combined_direction = g_value_get_uint (value);
-          new_sending = ((stream->combined_direction &
-                TP_MEDIA_STREAM_DIRECTION_SEND) != 0);
-          if (priv->sending != new_sending)
-            {
-              priv->sending = new_sending;
-              push_sending (stream);
-            }
-        }
+      stream->combined_direction = g_value_get_uint (value);
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -824,6 +815,16 @@ gabble_media_stream_new_native_candidate (GabbleMediaStream *self,
       1, transports,
       G_MAXUINT);
 
+  if (transports->len != 1)
+    {
+      GMS_DEBUG_WARNING (priv->session, "%s: number of transports was not 1; "
+          "rejecting", G_STRFUNC);
+      g_set_error (error, TELEPATHY_ERRORS, NotImplemented, "google p2p "
+          "connections only support the concept of one transport per "
+          "candidate");
+      return FALSE;
+    }
+
   transport = g_ptr_array_index (transports, 0);
   addr = g_value_get_string (g_value_array_get_nth (transport, 1));
   if (!strcmp (addr, "127.0.0.1"))
@@ -1091,25 +1092,19 @@ _gabble_media_stream_message_new (GabbleMediaStream *stream,
 {
   GabbleMediaStreamPrivate *priv = GABBLE_MEDIA_STREAM_GET_PRIVATE (stream);
   LmMessage *msg;
+  LmMessageNode *session_node = NULL;
 
   /* construct a session message */
   msg = _gabble_media_session_message_new (priv->session, action,
-                                           content_node);
+      &session_node);
 
-  /* add our content node to it if in jingle mode */
-  if (priv->mode == MODE_JINGLE)
-    {
-      LmMessageNode *node;
-
-      node = lm_message_node_add_child (*content_node, "content", NULL);
-      lm_message_node_set_attribute (node, "name", stream->name);
-
-      *content_node = node;
-    }
+  /* add our content node to it if necessary */
+  *content_node = _gabble_media_stream_add_content_node (stream, session_node);
 
   return msg;
 }
 
+
 static void
 push_candidate (GabbleMediaStream *stream, GValueArray *candidate)
 {
@@ -1640,6 +1635,28 @@ codec_params_from_tp_foreach (gpointer key, gpointer value, gpointer user_data)
       (priv->media_type == TP_CODEC_MEDIA_TYPE_AUDIO) ? "audio" : "video");
 }
 
+LmMessageNode *
+_gabble_media_stream_add_content_node (GabbleMediaStream *stream,
+                                       LmMessageNode *session_node)
+{
+  GabbleMediaStreamPrivate *priv = GABBLE_MEDIA_STREAM_GET_PRIVATE (stream);
+  LmMessageNode *node = session_node;
+
+  /* add our content node to it if in jingle mode */
+  if (priv->mode == MODE_JINGLE)
+    {
+      node = lm_message_node_add_child (session_node, "content", NULL);
+      lm_message_node_set_attribute (node, "name", stream->name);
+
+      if (priv->session->initiator == stream->initiator)
+        lm_message_node_set_attribute (node, "creator", "initiator");
+      else
+        lm_message_node_set_attribute (node, "creator", "responder");
+    }
+
+  return node;
+}
+
 void
 _gabble_media_stream_content_node_add_description (GabbleMediaStream *stream,
                                                    LmMessageNode *content_node)
@@ -1748,3 +1765,27 @@ _gabble_media_stream_content_node_add_transport (GabbleMediaStream *stream,
   return node;
 }
 
+void
+_gabble_media_stream_update_sending (GabbleMediaStream *stream,
+                                     gboolean start_sending)
+{
+  GabbleMediaStreamPrivate *priv;
+  gboolean new_sending;
+
+  g_assert (GABBLE_IS_MEDIA_STREAM (stream));
+
+  priv = GABBLE_MEDIA_STREAM_GET_PRIVATE (stream);
+
+  new_sending =
+    ((stream->combined_direction & TP_MEDIA_STREAM_DIRECTION_SEND) != 0);
+
+  if (priv->sending == new_sending)
+    return;
+
+  if (new_sending && !start_sending)
+    return;
+
+  priv->sending = new_sending;
+  push_sending (stream);
+}
+
