@@ -396,28 +396,12 @@ _google_roster_item_should_keep (LmMessageNode *item_node,
   if (!g_strdiff (attr, "true"))
     return FALSE;
 
-  /* skip email addresses the user has invited */
-  attr = lm_message_node_get_attribute (item_node, "gr:inv");
-
-  if (!g_strdiff (attr, "A"))
-    return FALSE;
-
   /* skip email addresses that replied to an invite */
   attr = lm_message_node_get_attribute (item_node, "gr:alias-for");
 
   if (attr != NULL)
     return FALSE;
 
-  /* skip hidden items */
-  if (item->google_type == GOOGLE_ITEM_TYPE_HIDDEN)
-    return FALSE;
-
-  /* allow items that have rejected a subscription */
-  attr = lm_message_node_get_attribute (item_node, "gr:rejected");
-
-  if (!g_strdiff (attr, "true"))
-    return TRUE;
-
   /* allow items that we've requested a subscription from */
   if (item->ask_subscribe)
     return TRUE;
@@ -498,8 +482,14 @@ _gabble_roster_item_update (GabbleRoster *roster,
     {
       item->google_type = _parse_google_item_type (node);
 
-      /* discard roster item if strange */
-      if (!_google_roster_item_should_keep (node, item))
+      /* discard roster item if strange, just hide it if it's hidden */
+      if (item->google_type == GOOGLE_ITEM_TYPE_HIDDEN)
+        {
+          DEBUG ("Google roster: caching hidden contact %d (%s)", handle,
+              lm_message_node_get_attribute (node, "jid"));
+          item->subscription = GABBLE_ROSTER_SUBSCRIPTION_NONE;
+        }
+      else if (!_google_roster_item_should_keep (node, item))
         {
           DEBUG ("Google roster: discarding odd contact %d (%s)", handle,
               lm_message_node_get_attribute (node, "jid"));
@@ -826,7 +816,7 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
   switch (sub_type)
     {
       LmMessageNode *item_node;
-      GIntSet *empty, *pub_add, *pub_rem,
+      GIntSet *pub_add, *pub_rem,
               *sub_add, *sub_rem, *sub_rp,
               *known_add, *known_rem,
               *deny_add, *deny_rem;
@@ -839,7 +829,6 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
     case LM_MESSAGE_SUB_TYPE_SET:
       /* asymmetry is because we don't get locally pending subscription
        * requests via <roster>, we get it via <presence> */
-      empty = g_intset_new ();
       pub_add = g_intset_new ();
       pub_rem = g_intset_new ();
       sub_add = g_intset_new ();
@@ -955,7 +944,10 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
             case GABBLE_ROSTER_SUBSCRIPTION_TO:
             case GABBLE_ROSTER_SUBSCRIPTION_FROM:
             case GABBLE_ROSTER_SUBSCRIPTION_BOTH:
-              g_intset_add (known_add, handle);
+              if (item->google_type == GOOGLE_ITEM_TYPE_HIDDEN)
+                  g_intset_add (known_rem, handle);
+              else
+                  g_intset_add (known_add, handle);
               break;
             case GABBLE_ROSTER_SUBSCRIPTION_REMOVE:
               g_intset_add (known_rem, handle);
@@ -996,21 +988,21 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
 
       DEBUG ("calling change members on publish channel");
       gabble_group_mixin_change_members (G_OBJECT (chan),
-            "", pub_add, pub_rem, empty, empty, 0, 0);
+            "", pub_add, pub_rem, NULL, NULL, 0, 0);
 
       handle = GABBLE_LIST_HANDLE_SUBSCRIBE;
       chan = _gabble_roster_get_channel (roster, handle);
 
       DEBUG ("calling change members on subscribe channel");
       gabble_group_mixin_change_members (G_OBJECT (chan),
-            "", sub_add, sub_rem, empty, sub_rp, 0, 0);
+            "", sub_add, sub_rem, NULL, sub_rp, 0, 0);
 
       handle = GABBLE_LIST_HANDLE_KNOWN;
       chan = _gabble_roster_get_channel (roster, handle);
 
       DEBUG ("calling change members on known channel");
       gabble_group_mixin_change_members (G_OBJECT (chan),
-            "", known_add, known_rem, empty, empty, 0, 0);
+            "", known_add, known_rem, NULL, NULL, 0, 0);
 
       if (google_roster)
         {
@@ -1019,7 +1011,7 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
 
           DEBUG ("calling change members on deny channel");
           gabble_group_mixin_change_members (G_OBJECT (chan),
-              "", deny_add, deny_rem, empty, empty, 0, 0);
+              "", deny_add, deny_rem, NULL, NULL, 0, 0);
 
           g_intset_destroy (deny_add);
           g_intset_destroy (deny_rem);
@@ -1029,7 +1021,6 @@ gabble_roster_iq_cb (LmMessageHandler *handler,
           _gabble_roster_item_remove (roster,
               g_array_index (removed, GabbleHandle, i));
 
-      g_intset_destroy (empty);
       g_intset_destroy (pub_add);
       g_intset_destroy (pub_rem);
       g_intset_destroy (sub_add);
@@ -1124,7 +1115,7 @@ gabble_roster_presence_cb (LmMessageHandler *handler,
   LmMessageNode *pres_node, *child_node;
   const char *from;
   LmMessageSubType sub_type;
-  GIntSet *empty, *tmp;
+  GIntSet *tmp;
   GabbleHandle handle;
   const gchar *status_message = NULL;
   GabbleRosterChannel *chan = NULL;
@@ -1173,16 +1164,14 @@ gabble_roster_presence_cb (LmMessageHandler *handler,
       DEBUG ("making %s (handle %u) local pending on the publish channel",
           from, handle);
 
-      empty = g_intset_new ();
       tmp = g_intset_new ();
       g_intset_add (tmp, handle);
 
       handle = GABBLE_LIST_HANDLE_PUBLISH;
       chan = _gabble_roster_get_channel (roster, handle);
       gabble_group_mixin_change_members (G_OBJECT (chan), status_message,
-          empty, empty, tmp, empty, 0, 0);
+          NULL, NULL, tmp, NULL, 0, 0);
 
-      g_intset_destroy (empty);
       g_intset_destroy (tmp);
 
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -1190,18 +1179,16 @@ gabble_roster_presence_cb (LmMessageHandler *handler,
       DEBUG ("removing %s (handle %u) from the publish channel",
           from, handle);
 
-      empty = g_intset_new ();
       tmp = g_intset_new ();
       g_intset_add (tmp, handle);
 
       handle = GABBLE_LIST_HANDLE_PUBLISH;
       chan = _gabble_roster_get_channel (roster, handle);
       changed = gabble_group_mixin_change_members (G_OBJECT (chan),
-          status_message, empty, tmp, empty, empty, 0, 0);
+          status_message, NULL, tmp, NULL, NULL, 0, 0);
 
       _gabble_roster_send_presence_ack (roster, from, sub_type, changed);
 
-      g_intset_destroy (empty);
       g_intset_destroy (tmp);
 
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -1209,18 +1196,16 @@ gabble_roster_presence_cb (LmMessageHandler *handler,
       DEBUG ("adding %s (handle %u) to the subscribe channel",
           from, handle);
 
-      empty = g_intset_new ();
       tmp = g_intset_new ();
       g_intset_add (tmp, handle);
 
       handle = GABBLE_LIST_HANDLE_SUBSCRIBE;
       chan = _gabble_roster_get_channel (roster, handle);
       changed = gabble_group_mixin_change_members (G_OBJECT (chan),
-          status_message, tmp, empty, empty, empty, 0, 0);
+          status_message, tmp, NULL, NULL, NULL, 0, 0);
 
       _gabble_roster_send_presence_ack (roster, from, sub_type, changed);
 
-      g_intset_destroy (empty);
       g_intset_destroy (tmp);
 
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -1228,18 +1213,16 @@ gabble_roster_presence_cb (LmMessageHandler *handler,
       DEBUG ("removing %s (handle %u) from the subscribe channel",
           from, handle);
 
-      empty = g_intset_new ();
       tmp = g_intset_new ();
       g_intset_add (tmp, handle);
 
       handle = GABBLE_LIST_HANDLE_SUBSCRIBE;
       chan = _gabble_roster_get_channel (roster, handle);
       changed = gabble_group_mixin_change_members (G_OBJECT (chan),
-          status_message, empty, tmp, empty, empty, 0, 0);
+          status_message, NULL, tmp, NULL, NULL, 0, 0);
 
       _gabble_roster_send_presence_ack (roster, from, sub_type, changed);
 
-      g_intset_destroy (empty);
       g_intset_destroy (tmp);
 
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -1581,3 +1564,40 @@ gabble_roster_handle_remove (GabbleRoster *roster,
 
   return ret;
 }
+
+gboolean
+gabble_roster_handle_add (GabbleRoster *roster,
+                          GabbleHandle handle,
+                          GError **error)
+{
+  GabbleRosterPrivate *priv = GABBLE_ROSTER_GET_PRIVATE (roster);
+  GabbleRosterItem *item;
+  LmMessage *message;
+  gboolean do_add = FALSE;
+  gboolean ret;
+
+  g_return_val_if_fail (roster != NULL, FALSE);
+  g_return_val_if_fail (GABBLE_IS_ROSTER (roster), FALSE);
+  g_return_val_if_fail (gabble_handle_is_valid (priv->conn->handles,
+      TP_HANDLE_TYPE_CONTACT, handle, NULL), FALSE);
+
+  if (!gabble_roster_handle_has_entry (roster, handle))
+      do_add = TRUE;
+
+  item = _gabble_roster_item_get (roster, handle);
+
+  if (item->google_type == GOOGLE_ITEM_TYPE_HIDDEN)
+    {
+      item->google_type = GOOGLE_ITEM_TYPE_NORMAL;
+      do_add = TRUE;
+    }
+
+  if (!do_add)
+      return TRUE;
+
+  message = _gabble_roster_item_to_message (roster, handle, NULL);
+  ret = _gabble_connection_send (priv->conn, message, error);
+  lm_message_unref (message);
+
+  return ret;
+}
